function [C,U,V,W]=BKNonSym(A,U0,V0,W0,p,stage)
% Nonsymmetric BK: Symmetric Block Krylov applied to B=[0 A; A' 0] implicitly
% A is a 3-tensor, nonsymmetric
% U0, V0, W0 have second dimension r1, r2, r3
% p : Block size
% stage : number of stages
% Output : 
% C : Core tensor 
% U=[U1 U2 ...] (as many as are generated in the stages executed 
% Same for V and W 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Distributed under a 3-Clause BSD License 
%
% Copyright 2021 Lars Eldén, Linköping University, and Maryam Dehghan,
% Federal University of Ceará,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if stage>3, 
    disp('Stage 3 not implemented, stage:=3')
    stage=3;
end

n=size(A);

%%% Stage 1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[U,U1]=newblock(A,U0,V0,W0,[2 3],1);

[V,V1]=newblock(A,V0,U0,W0,[1 3],2);

[W,W1]=newblock(A,W0,U0,V0,[1 2],3);

%%%  Stage 2  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if stage > 1 
    if size(U,2)<n(1)
        % U2  <-- V0,W1
        [U,U2]=newblock(A,U,V0,barX(W1,p),[2 3],1);
        if size(U,2)<n(1)
            % U3 <-- V1,W0
            [U,U3]=newblock(A,U,barX(V1,p),W0,[2 3],1);
        end
    end
    
    if size(V,2)<n(2)
        % V2 <-- U0,W1
        [V,V2]=newblock(A,V,U0,barX(W1,p),[1 3],2);
        if size(V,2)<n(2)
            % V3 <-- U1,W0
            [V,V3]=newblock(A,V,barX(U1,p),W0,[1 3],2);
        end
    end
    
    if size(W,2)<n(3)
        % W2 <-- U0,V1
        [W,W2]=newblock(A,W,U0,barX(V1,p),[1 2],3);
        if size(W,2)<n(3)
            % W3 <-- U1,V0
            [W,W3]=newblock(A,W,barX(U1,p),V0,[1 2],3);
        end
    end
    
    if stage==3
    %%% stage 3 simplified %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        if size(U,2)<n(1) & exist('W2') 
            % U4 <-- V0,W2 
            [U,U4]=newblock(A,U,V0,barX(W2,p),[2 3],1);
            if size(U,2)<n(1)  & exist('V2')
                % U5 <-- V2,W0
                [U,U5]=newblock(A,U,barX(V2,p),W0,[2 3],1);
            end
        end
        if size(V,2)<n(2) & exist('W2')
            % V4 <-- U0,W2
            [V,V4]=newblock(A,V,U0,barX(W2,p),[1 3],2);
            if size(V,2)<n(2) & exist('U2')
                % V5 <-- U2,W0
                [V,V5]=newblock(A,V,barX(U2,p),W0,[1 3],2);
            end
        end
        if size(W,2)<n(3) & exist('V2')
            % W5 <-- U0,V2
            [W,W5]=newblock(A,W,U0,barX(V2,p),[1 2],3);
            if size(W,2)<n(3) & exist('U2')
                % W6= <-- U2,V0
                [W,W6]=newblock(A,barX(U2,p),V0,[1 2],3);
            end
        end        
    end % stage=3
end % stage > 1
C=ttm(A,{U,V,W},'t');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Xbar=barX(X,p)
if size(X,2)<p
    Xbar=X;
else
    Xbar=X(:,1:p);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Xall,Xnew]=newblock(A,X,Y,Z,modeYZ,modeX)
Xhat=tam(double(ttm(A,{Y,Z},modeYZ,'t')),modeX);
%[Xhat,rhat,perm]=qr(Xhat,0);  % Permute columns to make the first columns of Xnew
                              % the most significant of the new.
                              % Initial tests show that this does not have
                              % any noticeable impact on the convergence rate.
k=size(X,2);
n=size(A);
[Xall,rall]=qr([X Xhat],0);   % No pivoting here
if size(Xall,2)>n(modeX)      % Check if Xall has full modeX dimension
    Xall=Xall(:,1:n(modeX));
end
Xnew=Xall(:,k+1:end);




