function [Iu,Iv,Iw,UIu,VIv,WIw,U,V,W,OutConv,F]=BlockKrylovSchurNonSym(A,options)
% Krylov-Schur for nonsymmetric Tensor %%%%%%%%%%%%%%%%%%%%%
%
% A  Sparse 3-tensor that is nonsymmetric
%
% Method: BKS: Block Krylov-Schur for computing the best rank-(r) approximation 
% Options  r:     rank-(r_1,r_2,r_3)
%          stage: number of stages   
%          p:     block size
% Outdata 
%         Iu, Iv, Iw   : Reorderings in modes 1,2 and 3
%         UIu, VIv, WIw: Reordered solution 
%         U, V, W      : Solution, (not reordered)
%         OutConv      : Relative gradient as a function of iterations
%         F            : Core tensor of dimension r

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Distributed under a 3-Clause BSD License 
%
% Copyright 2021 Lars Eldén, Linköping University, and Maryam Dehghan,
% Federal University of Ceará,
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%% More options %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if isfield(options,'r'),r=options.r; else r=[2 2 2]; end
if isfield(options,'p'),p=options.p; else p=4; end
if isfield(options,'stage'),stage=options.stage; else stage=2; end
if isfield(options,'method'), method=options.method; else method=0; end
if isfield(options,'initHOOI'),initHOOI=options.initHOOI;else,initHOOI=5;end % HOOI for outer starting value
if isfield(options,'innerHOOI'),innerHOOI=options.innerHOOI;else,innerHOOI=5;end % HOOI for inner starting value
if isfield(options,'tolout'),tolout=options.tolout;else,tolout=1e-13;end     % Outer stopping criterion
if isfield(options,'maxitNG'),maxitNG=options.maxitNG;else,maxitNG=20;end    % Inner maximum iterations
if isfield(options,'tolin'),tolin=options.tolin;else,tolin=1e-13;end         % Inner stopping criterion
if isfield(options,'maxit'), maxit=options.maxit; else maxit=50; end
if isfield(options,'BKSverbose'), BKSverbose=options.BKSverbose; else BKSverbose=1; end
if isfield(options,'NGverbose'), NGverbose=options.NGverbose; else NGverbose=1; end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Dimensions 

if BKSverbose
    disp('------------------------')
    n=size(A); disp(['Tensor dimensions:  ' num2str(n(1)) ' x ' num2str(n(2)) ' x ' num2str(n(3))]);
    disp(['BKS for r = ' num2str(r) '  Parameters, stage, p:  ' num2str([stage p])])
    disp('------------------------')
end

%% Krylov_Schur  First starting value from HOOI with r=[1 1 1]

% Random starting point of HOOI method
for i=1:3
    X0_HOOI{i}=orth(rand([n(i),1]));
end

% HOOI for r=[1 1 1]
[XX1, f1, nn1] = algALS(A,X0_HOOI,[1 1 1]',initHOOI,tolout,initHOOI+2);  % itstep=maxitHOOI+2: 
                                                                           % Do not compute the gradient 

%% BKS method

uStart=XX1{1};
vStart=XX1{2};
% Use BLOCK-Krylov to get (k_1,k_2,k_3) spaces
[Core_Tensor,U,V,W]=BKNonSymFirst(A,r,uStart,vStart,p,stage);
k=size(Core_Tensor); k=k';
if BKSverbose
    disp('------------------------')
    disp(['Dimensions of expanded core tensor:   ' num2str(k')])
    disp('------------------------')
end

it=0;
done=0;

% Krylov-Schur loop
while ~done && it < maxit
    it=it+1;
    %%%%%%%%%%%%%%%%%%%%%%% Compute  the (r,r,r)-rank approximation of core tensor by NG method %%%%%%%%%%%
    
    [C_svd,U_svd,V_svd,W_svd]=hosvd_Truncate_r_nonSym(Core_Tensor,r);  % HOSVD starting approximation
    
    if innerHOOI>0                                 % Initial AOI iterations
        X0{1}=U_svd; X0{2}=V_svd; X0{3}=W_svd;
        [XX11,f11,nn11]=algALS(Core_Tensor,X0,r',innerHOOI,tolin);
    else
        XX11{1}=U_svd;XX11{2}=V_svd;XX11{3}=W_svd;
    end
    
    [XX,ff,nn2]=algNlc(Core_Tensor,XX11,k,r,maxitNG,tolin,NGverbose); % Newton-Grassmann
    UXX1=U*XX{1};VXX2=V*XX{2};
    WXX3=W*XX{3};
    XXX{1}=UXX1; XXX{2}=VXX2;XXX{3}=WXX3;    % New approximation
    
    % Check convergence of outer iteration (K-S) for new approximation     
    % First do the expansion for next NG step
 
    [Core_Tensor,U,V,W]=BKNonSym(A,UXX1,VXX2,WXX3,p,stage); 
     
    % Compute gradient using the new Core_Tensor (Proposition 5.1 and figure)
    [F,normPhi(it),normGrad(it)]=norm_phi_gradientLocC(Core_Tensor,r);
    relative_gradient=normGrad(it)/normPhi(it);
    if relative_gradient<tolout
        done=1;
    end
end
% After convergence 
OutConv=normGrad./normPhi;   % Relative gradient
U=U(:,1:r(1)); V=V(:,1:r(2)); W=W(:,1:r(3)); % Remove the last expansion

if BKSverbose
    disp('------------------------')
    disp(['BKS, stage, p, k:  ' num2str(stage) '  ' num2str(p) '  '...
        num2str(k')])
    disp('------------------------')
end
 
%% Sort and reorder
% Make sure that U(:,2) is the one with the sign change
if max(U(:,1))*min(U(:,1))>=0
    aaaaaa=1;     % Do nothing
elseif max(U(:,2))*min(U(:,2))>=0,
    U(:,1:2)=U(:,2:-1:1);   
end
[uu,Iu]=sort(-U(:,2));

% Make sure that V(:,2) is the one with the sign change
if max(V(:,1))*min(V(:,1))>=0
    aaaaaa=1;     % Do nothing
elseif max(V(:,2))*min(V(:,2))>=0,
    V(:,1:2)=V(:,2:-1:1);   
end
[vv,Iv]=sort(-V(:,2));

if r(3)>1
  if max(W(:,1))*min(W(:,1))>=0
    aaaaaa=1;     % Do nothing
elseif max(W(:,2))*min(W(:,2))>=0,
    W(:,1:2)=W(:,2:-1:1);   
end
 [ww,Iw]=sort(W(:,2));
else
    Iw=1:size(W,1);
end

UIu=U(Iu,:);
VIv=V(Iv,:);
WIw=W(Iw,:);
F=ttm(A,{U,V,W},'t');


     
 
